/*******************************************************************************
 *                                                                             *
 *  Filename : octree.c                                                        *
 *  Author   : Bruno Grossniklaus                                              *
 *  Created  : 22.4.91                                                         *
 *  Modified :                                                                 *
 *                                                                             *
 ******************************************************************************/

/*
 * octree.c
 *
 * Transforms 32 or 24 Bit pixrect or rayshade files to 8 Bit colormap files
 * with the octree algorithm.
 *
 */


/******************************************************************** INCLUDE */
/* global defaults are set in defaults.h */
#include <stdio.h>
/* This version is for use with XView only! */
#include <pixrect/pixrect_hs.h>
#include "defaults.h"


/******************************************************************** DEFINE  */
/* defaults for popular */
/* more defaults are included in defaults.h */

#define VERSION "1.02 (20.6.91)"          /* Versions Nr.                     */

#define MAXDEPTH       8                  /* maximum depth of the octree      */
#define MAPSIZE       256                 /* maximum of colormap colors       */

#define DEFAULT_OPT_SPEED   TRUE          /* Default optimize.                */
#define DEFAULT_OPT_ERROR   FALSE         /* ONLY ONE OF THEM CAN BY TRUE     */
#define DEFAULT_OPT_DETAIL  FALSE


/********************************************************************* TYPES  */
/* types for octree */
struct NODE {                             /* the node                         */
  unsigned char       level;              /* level of node                    */
  unsigned char       leaf;               /* is node leaf or not              */
  struct NODE         *next[8];           /* pointers to next nodes           */
  unsigned long int   color_count;        /* to count the colors in node      */
  int                 color_index;        /* for faster mapping               */
  unsigned long int   r,g,b;              /* !sum of colors in node           */
  struct NODE         *next_node;         /* for faster access in ReduceTree  */
};


/******************************************************************** GLOBALS */
static int            global_size = 0;    /* number of leaves                 */
static unsigned char  global_octree_depth = 8; /* depth of the octree         */
static unsigned int   global_map_index = 0; /* for mapping colors             */
static unsigned char  global_red[256];    /* used for colormap                */
static unsigned char  global_green[256];  /* used for colormap                */
static unsigned char  global_blue[256];   /* used for colormap                */
static struct NODE    *global_reduce_list[MAXDEPTH]; /* for faster access     */

static unsigned char  flag_opt_speed  = DEFAULT_OPT_SPEED;  /* optimize method*/
static unsigned char  flag_opt_error  = DEFAULT_OPT_ERROR;
static unsigned char  flag_opt_detail = DEFAULT_OPT_DETAIL;


/****************************************************************** FUNCTIONS */
/* functions in pic_io.c */
extern void set_file_names();
extern void close_success();
extern void close_error();
extern int  open_in_file();
extern int  open_out_file();
extern int  get_image();
extern int  get_mem();
extern int  put_image();
extern void print_euclidean_dist();
extern void write_map();


/*******************************************************************************
 *   main                                                                      *
 *                                                                             *
 *   Input  :                                                                  *
 *            - argv      arguments (options and filename)                     *
 *            - argc      number of arguments                                  *
 *                                                                             *
 *   Return : - NULL if ok, 1 on error                                         *
 *                                                                             *
 ******************************************************************************/
int main (argc,argv)
int argc;
char *argv [];

{
/******************************************************************** DEKLAR  */
  FILE                *fp_in;             /* file pointer                     */
  FILE                *fp_out;            /* file pointer                     */
  char                fn_in[FNS];         /* file name input                  */
  char                fn_out[FNS];        /* file name output                 */
  char                tmp_str[16];        /* temorary string                  */
  char                tmp_str_fs[3*FNS];  /* for system call fs               */

  Pixrect             *mem_8_out;         /* pixrect in memory 8 Bit          */
  Pixrect             *mem_in;            /* pixrect in memory 24 or 32 Bit   */

  colormap_t          colormap;           /* colormap of output file          */
  unsigned int        map_size = MAPSIZE; /* size of colormap                 */

  unsigned char       *im_ptr_8_out;      /* pointer into 8 Bit image         */
  int                 *im_ptr_in;         /* pointer into 24 or 32 Bit image  */

  int                 x,y,pic_depth;      /* size and depth of image          */
  int                 out_format;         /* outfileformat (standard,encoded) */
  int                 i;                  /* used for loops                   */

  unsigned char       flag_prompt;        /* flag if prompt                   */
  unsigned char       flag_euclid;        /* flag if calculate euclid         */
  unsigned char       flag_expand_name;   /* flag if expand name              */
  unsigned char       flag_option_ok;     /* flag if option ok                */
  unsigned char       flag_write_map;     /* flag if write colormap           */
  unsigned char       flag_stdout;        /* flag if write to stdout          */
  unsigned char       flag_sunview;       /* flag if sunview                  */

  struct NODE         *main_tree;         /* the root of the octree           */
  unsigned char       r,g,b;              /* color values                     */

  unsigned char       flag_floyd_steinb;  /* flag if floyd steinberg dither   */
  register int                            /* spreading of errors              */
                      spread_hori,        /* spreading horizontal             */
                      spread_diag,        /* spreading diagonal               */
                      spread_vert;        /* spreading vertical               */
  register int        spread_percent;     /* spread error in %                */

/******************************************************************** INIT    */
  /* set variables with defaults (defined in defaults.h) */

  /* file names with EOS (end of string) */
  for (i = 0; i < FNS; i++) fn_in[i] = fn_out[i] = '\0';

  /* pointers with NULL */
  fp_in = fp_out = NULL;
  mem_8_out = mem_in = NULL;

  /* other default values */
  out_format = DEFAULT_OUT_FORM;
  flag_prompt = DEFAULT_PROMPT;
  flag_euclid = DEFAULT_EUCLID;
  flag_expand_name = DEFAULT_EXP_NAM;
  flag_write_map = DEFAULT_WRITE_MAP;
  flag_stdout = FALSE;

  /* set default for floyd steinberg dither */
  flag_floyd_steinb = DEFAULT_FLOYD;
  spread_hori = DEFAULT_SPREAD_HORI;
  spread_diag = DEFAULT_SPREAD_DIAG;
  spread_vert = 100 - spread_hori - spread_diag;
  spread_percent = DEFAULT_SPREAD_PERCENT;

  fprintf (stderr,"\n");


/******************************************************************** OPTIONS */
  /* argv[0] is name of program */
  /* if there are no parameters specified show usage */
  if (argc <= 1) {
    fprintf (stderr,"Called without parameters \n");
    usage_exit (argv[0]);
  }

  /* the last parameter is the input file name or the -h option */
  else if (argv[argc - 1] [0] != '-')
      strcpy (fn_in,argv[argc - 1]);
    else
      if (strcmp (argv[argc - 1], "-h") == 0) help_exit (argv[0]);
      else if (strcmp (argv[argc - 1],"-c") == 0) {
        fprintf (stderr,"%s was written by Bruno Grossniklaus in 1991\n",
                         argv[0]);
        exit (1);
      }
      else {
        fprintf (stderr,"Last parameter %s is not filename (but option) \n",
                                                                  argv[argc-1]);
        usage_exit (argv[0]);
      }


  /* scan the options */
  i=1;
  flag_option_ok = TRUE;

  while ( (i < argc - 1) && (flag_option_ok) ) {
    flag_option_ok = FALSE;


    if (strcmp (argv[i],"-p+") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_prompt = TRUE;
    }

    if (strcmp (argv[i],"-p-") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_prompt = FALSE;
    }

    if (strcmp (argv[i],"-p") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_prompt = !flag_prompt;
    }


    if (strcmp (argv[i],"-o") == 0) {
      flag_option_ok = TRUE;
      i++;
      if (i >= argc - 1) {
        fprintf (stderr,"Missing parameter after %s. \n",argv[i-1]);
        fprintf (stderr,"%s is the input filename! \n",argv[i]);
        usage_exit (argv[0]);
      }
      strcpy (fn_out,argv[i]);
      if ( (fn_out[0] == '-') || (fn_out == NULL) ) {
        fprintf (stderr,"Parameter %s after -o is not filename \n",fn_out);
        usage_exit (argv[0]);
      }
      i++;
    }


    if (strcmp (argv[i],"-en+") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_expand_name = TRUE;
    }

    if (strcmp (argv[i],"-en-") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_expand_name = FALSE;
    }

    if (strcmp (argv[i],"-en") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_expand_name = !flag_expand_name;
    }


    if ( (strcmp (argv[i],"-ws+") == 0) || (strcmp (argv[i],"-ws") == 0) ) {
      flag_option_ok = TRUE;
      out_format = RT_STANDARD;
      if (flag_prompt) fprintf (stderr,"Output file format is STANDARD \n");
      i++;
    }

    if (strcmp (argv[i],"-ws-") == 0) {
      flag_option_ok = TRUE;
      out_format = RT_BYTE_ENCODED;
      if (flag_prompt)
            fprintf (stderr,"Output file format is RUN LENGTH ENCODED \n");
      i++;
    }


    if (strcmp (argv[i],"-sv+") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_sunview = TRUE;
    }

    if (strcmp (argv[i],"-sv-") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_sunview = FALSE;
    }

    if (strcmp (argv[i],"-sv") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_sunview = !flag_sunview;
    }


    if (strcmp (argv[i],"-opt") == 0) {
      i++;

      if (strcmp (argv[i],"speed") == 0) {
        i++;
        flag_option_ok = TRUE;
        flag_opt_speed  = TRUE;
        flag_opt_error  = FALSE;
        flag_opt_detail = FALSE;
      }

      else if (strcmp (argv[i],"error") == 0) {
        i++;
        flag_option_ok = TRUE;
        flag_opt_speed  = FALSE;
        flag_opt_error  = TRUE;
        flag_opt_detail = FALSE;
      }

      else if (strcmp (argv[i],"detail") == 0) {
        i++;
        flag_option_ok = TRUE;
        flag_opt_speed  = FALSE;
        flag_opt_error  = FALSE;
        flag_opt_detail = TRUE;
      }
    }

    if (strcmp (argv[i],"-fs%") == 0) {
      i++;
      flag_option_ok = TRUE;
      if ( (argv[i][0] == '-') || (i >= argc - 1) ) {
        fprintf (stderr,"Missing parameter after %s. \n",argv[i-1]);
        usage_exit (argv[0]);
      }
      else {
        spread_percent = atoi (argv[i]);
        if ( (spread_percent < 0) || (spread_percent > 100) ) {
          fprintf (stderr,"After -fs%% must be a number in percent. ");
          fprintf (stderr,"%d not allowed\n",spread_percent);
          usage_exit (argv[0]);
        }
      }
      i++;
    }

    if (strcmp (argv[i],"-fs+") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_floyd_steinb = TRUE;
      if ( (argv[i][0] != '-')  && (i < argc - 1) ) {
        spread_hori = atoi (argv[i]);
        i++;
        if ( (argv[i][0] == '-') || (i >= argc - 1) ) {
          fprintf (stderr,"Missing parameter after %s. \n",argv[i-2]);
          fprintf (stderr,"%s must be diagonal spreading! \n",argv[i]);
          usage_exit (argv[0]);
        }
        spread_diag = atoi (argv[i]);
        i++;
      }
      if ( (spread_hori < 0) || (spread_hori > 100) ) {
        fprintf (stderr,"Horizontal spreading is ");
        fprintf (stderr,"%d, but must be >= 0 and <= 100!\n",
                 spread_hori);
        usage_exit (argv[0]);
      }
      if ( (spread_diag < 0) || (spread_diag > 100) ) {
        fprintf (stderr,"Diagonal spreading is ");
        fprintf (stderr,"%d, but must be >= 0 and <= 100!\n",
                 spread_diag);
        usage_exit (argv[0]);
      }
      spread_vert = 100 - spread_hori - spread_diag;
      if (spread_vert < 0) {
        fprintf (stderr,"Vertical spreading is ");
        fprintf (stderr,"%d, but must be >= 0!\n",
                 spread_vert);
        usage_exit (argv[0]);
      }
    }

    if (strcmp (argv[i],"-fs-") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_floyd_steinb = FALSE;
    }

    if (strcmp (argv[i],"-fs") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_floyd_steinb = !flag_floyd_steinb;
      if (flag_floyd_steinb) {
        if ( (argv[i][0] != '-') && (i < argc - 1) ) {
          spread_hori = atoi (argv[i]);
          i++;
          if ( (argv[i][0] == '-') || (i >= argc - 1) ) {
            fprintf (stderr,"Missing parameter after %s. \n",argv[i-2]);
            fprintf (stderr,"%s must be diagonal spreading! \n",argv[i]);
            usage_exit (argv[0]);
          }
          spread_diag = atoi (argv[i]);
          i++;
        }
        if ( (spread_hori < 0) || (spread_hori > 100) ) {
          fprintf (stderr,"Horizontal spreading is ");
          fprintf (stderr,"%d, but must be >= 0 and <= 100!\n",
                   spread_hori);
          usage_exit (argv[0]);
        }
        if ( (spread_diag < 0) || (spread_diag > 100) ) {
          fprintf (stderr,"Diagonal spreading is ");
          fprintf (stderr,"%d, but must be >= 0 and <= 100!\n",
                   spread_diag);
          usage_exit (argv[0]);
        }
        spread_vert = 100 - spread_hori - spread_diag;
        if (spread_vert < 0) {
          fprintf (stderr,"Vertical spreading is ");
          fprintf (stderr,"%d, but must be >= 0!\n",
                   spread_vert);
          usage_exit (argv[0]);
        }
      }
    }

    if (strcmp (argv[i],"-ce+") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_euclid = TRUE;
    }

    if (strcmp (argv[i],"-ce-") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_euclid = FALSE;
    }

    if (strcmp (argv[i],"-ce") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_euclid = !flag_euclid;
    }


    if (strcmp (argv[i],"-h") == 0)
      help_exit (argv[0]);


    if (strcmp (argv[i],"-m+") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_write_map = TRUE;
    }

    if (strcmp (argv[i],"-m-") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_write_map = FALSE;
    }

    if (strcmp (argv[i],"-m") == 0) {
      i++;
      flag_option_ok = TRUE;
      flag_write_map = !flag_write_map;
    }
  }


  /* if current parameter was no good option */
  if (!flag_option_ok) {
    fprintf (stderr,"%s is not an allowed option \n",argv[i]);
    usage_exit (argv[0]);
  }


/******************************************************************** INIT    */
  if (strcmp (fn_out,DEFAULT_STDOUT) == 0) flag_stdout = TRUE;

  /* function in pic_io.c */
  set_file_names (fn_in,fn_out);

  /* if expand name is set then add current settings to the output file name */
  if ( (flag_expand_name) && (!flag_stdout) ) {

    /* o means octree algorithm */
    strcat (fn_out,"_o");

    /* add optimization */
    if (flag_opt_speed)  strcat (fn_out,"_speed");
    if (flag_opt_error)  strcat (fn_out,"_error");
    if (flag_opt_detail) strcat (fn_out,"_detail");

    /* print the output file name to console */
    fprintf (stderr,"Expand name is set. So output file is %s \n",fn_out);
  }

  if ( (flag_stdout) && (flag_floyd_steinb)) strcpy (fn_out,TMP_FILE_NAME);

 /**************************************************************** OPEN & GET */
  /* now open the files and load image */
  /* functions in pic_io.c */
 if (open_in_file (fn_in,&fp_in,fp_out,mem_in,mem_8_out) != NULL) exit (1);
  if (open_out_file (fn_out,fp_in,&fp_out,mem_in,mem_8_out) != NULL) exit (1);
  if (    get_image (fn_in,fp_in,fp_out,&mem_in,mem_8_out,
                     flag_prompt,&x,&y,&pic_depth) != NULL) exit (1);

  /* allocate memory for output image */
  if (get_mem (fp_in,fp_out,mem_in,&mem_8_out,flag_prompt,x,y)
      != NULL) exit (1);


 /**************************************************************** BUILD TREE */
  global_size = 0;
  main_tree = NULL;

  /* if sunview option reserve 2 colors for sunview */
  if (flag_sunview) {
    map_size = map_size - 2;
   }

  /* init reduce list */
  for (i = 0; i < MAXDEPTH; i++) global_reduce_list[i] = NULL;

  if (flag_prompt) {
    fprintf (stderr,"Building tree ");
    /* only one of them is TRUE */
    if (flag_opt_speed)  fprintf (stderr,"(optimize speed) ... ");
    if (flag_opt_error)  fprintf (stderr,"(optimize error) ... ");
    if (flag_opt_detail) fprintf (stderr,"(optimize detail) ... ");
  }

  /* set pointer to first pixel */
  im_ptr_in = (int *) mpr_d(mem_in)->md_image;

  /* loop over all pixels in x direction */
  for (i = 0; i < x*y; i++, im_ptr_in++) {

      r =   (*im_ptr_in)        & 0xff;
      g =  ((*im_ptr_in) >>  8) & 0xff;
      b =  ((*im_ptr_in) >> 16) & 0xff;

      /* put color in tree */
      InsertTree (&main_tree,r,g,b,1);

      /* reduce down to map_size colors */
      while (global_size > map_size) ReduceTree();
  }

  if (flag_prompt) fprintf (stderr,"done. \n");


 /************************************************************ BUILD COLORMAP */
  if (flag_prompt) fprintf (stderr,"Choosing colormap ... ");

  global_map_index = 0;
  for (i = 0; i < 256; i++)
                       global_red[i] = global_green[i] = global_blue[i] = 0;

  InitColorTable (main_tree);

  colormap.type = RMT_EQUAL_RGB;
  colormap.length = 256;

  colormap.map[0] = global_red;
  colormap.map[1] = global_green;
  colormap.map[2] = global_blue;

  if (flag_sunview) {
    global_red[254] = global_green[254] = global_blue[254] = 255;
    global_red[255] = global_green[255] = global_blue[255] = 0;
  }

  /* fix bug or undocumented function of pixrect functions */
  if ( (global_red[0] == 0) && (global_green[0] == 0) &&
       (global_blue[0] == 0) ) global_blue[0] = 1;

  if (flag_prompt) fprintf (stderr,"done. \n");
  if (flag_prompt) fprintf (stderr,"There are %d colors in the colormap. \n",
                                    global_map_index);


/************************************************************* MAPPING COLORS */
  /* map colors to colormap */
  if (flag_prompt) fprintf (stderr,"Mapping colors ... ");

  /* set image pointers to first pixel */
  im_ptr_in = (int *) mpr_d(mem_in)->md_image;
  im_ptr_8_out = (unsigned char *) mpr_d(mem_8_out)->md_image;

  /* loop over all pixels in x direction */
  for (i = 0; i < x*y; i++, im_ptr_in++, im_ptr_8_out++) {

      r =   (*im_ptr_in)        & 0xff;
      g =  ((*im_ptr_in) >>  8) & 0xff;
      b =  ((*im_ptr_in) >> 16) & 0xff;

      *im_ptr_8_out = Quant (main_tree,r,g,b);
    }
  if (flag_prompt) fprintf (stderr,"done. \n");



/**************************************************************** WRITE IMAGE */
  /* write output image */
  if (put_image (fn_out,fp_in,fp_out,mem_in,mem_8_out,
      flag_prompt,colormap,out_format) != NULL) exit (1);


/************************************************************ WRITE TO STDOUT */
  /* write colormap to stdout if option is set */
  if (flag_write_map)
    if (flag_stdout) {
      fprintf (stderr,"I write the image to stdout. ");
      fprintf (stderr,"So no other writing to stdout allowed! \n");
    }
    else write_map (fn_out,global_red,global_green,global_blue);

  /* calculate euclidean distance if option is set */
  if (flag_euclid) 
    if (flag_stdout) {
      fprintf (stderr,"I write the image to stdout. ");
      fprintf (stderr,"So no other writing to stdout allowed! \n");
    }
    else print_euclidean_dist (fn_out,mem_in,mem_8_out,flag_prompt,
                               global_red,global_green,global_blue,x,y);


/**************************************************************** ENDING MAIN */
  /* close files and free memory */
  free ( (char *) main_tree);
  close_success (fp_in,fp_out,mem_in,mem_8_out);


  /* call fs from shell if fs option is set */
  if (flag_floyd_steinb) {
    strcpy  (tmp_str_fs,"fs -fs ");
    sprintf (tmp_str,"%d",spread_hori);
    strcat  (tmp_str_fs,tmp_str);
    strcat  (tmp_str_fs," ");
    sprintf (tmp_str,"%d",spread_diag);
    strcat  (tmp_str_fs,tmp_str);
    strcat  (tmp_str_fs," -fs% ");
    sprintf (tmp_str,"%d",spread_percent);
    strcat  (tmp_str_fs,tmp_str);
    if (flag_prompt)
      strcat (tmp_str_fs," -p+");
    else
      strcat (tmp_str_fs," -p-");
    if (flag_expand_name)
      strcat (tmp_str_fs," -en+");
    else
      strcat (tmp_str_fs," -en-");
    if (flag_euclid)
      strcat (tmp_str_fs," -ce+");
    else
      strcat (tmp_str_fs," -ce-");
    if (out_format == RT_STANDARD)
      strcat (tmp_str_fs," -ws+");
     else
      strcat (tmp_str_fs," -ws-");
    if (flag_stdout) {
      strcat  (tmp_str_fs," -o ");
      strcat  (tmp_str_fs,DEFAULT_STDOUT);
      strcat  (tmp_str_fs," ");
      strcat  (tmp_str_fs,fn_in);
      strcat  (tmp_str_fs," ");
      strcat  (tmp_str_fs,TMP_FILE_NAME);
    }
    else {
      strcat  (tmp_str_fs," ");
      strcat  (tmp_str_fs,fn_in);
      strcat  (tmp_str_fs," ");
      strcat  (tmp_str_fs,fn_out);
    }


    if (system (tmp_str_fs) != NULL) {
      fprintf (stderr,"Error in systemcall for fs!\n");
      fprintf (stderr," - maybe there was an error in fs.\n");
      fprintf (stderr," - maybe fs not in current dir or path.\n");
      fprintf (stderr," - maybe out of memory.\n");
      fprintf (stderr," - maybe system problem.\n");
      exit (1);
    }  
  }

  /* null problemo */
  return (0);
}


/*******************************************************************************
 *   Bit                                                                       *
 *                                                                             *
 *   Input  :                                                                  *
 *            - pos     take Bit in pos                                        *
 *            - number  number to get Bit                                      *
 *                                                                             *
 *   Output :                                                                  *
 *            - returns the Bit in position pos of number                      *
 ******************************************************************************/
unsigned char Bit (pos, number)
unsigned char pos;
unsigned char number;

{
  return ( (unsigned char) (number << (7-pos)) >> 7);
}


/*******************************************************************************
 *   Branch                                                                    *
 *                                                                             *
 *   Input  :                                                                  *
 *            - loc_r      color value                                         *
 *            - loc_g      color value                                         *
 *            - loc_b      color value                                         *
 *            - loc_depth  local depth                                         *
 *                                                                             *
 *   Output :                                                                  *
 *            - returns the branch of the octree for the color r g b in depth  *
 *              loc_depth.                                                     *
 ******************************************************************************/
unsigned char Branch (loc_r, loc_g, loc_b, loc_depth)
unsigned char  loc_r, loc_g, loc_b;
unsigned char  loc_depth;

{
  return (Bit (MAXDEPTH - loc_depth, loc_r) * 4 +
          Bit (MAXDEPTH - loc_depth, loc_g) * 2 +
          Bit (MAXDEPTH - loc_depth, loc_b) );
}


/*******************************************************************************
 *   InsertTree                                                                *
 *                                                                             *
 *   Input  :                                                                  *
 *            - loc_tree  pointer to pointer of tree                           *
 *            - loc_r     color to insert                                      *
 *            - loc_g     color to insert                                      *
 *            - loc_b     color to insert                                      *
 *            - loc_depth we are in depth loc_depth                            *
 *                                                                             *
 *   Output :                                                                  *
 *            - loc_tree  points to new node                                   *
 *                                                                             *
 *            inserts the color loc_r, loc_g, loc_b into the                   *
 *            subtree loc_tree in depth loc_depth                              *
 ******************************************************************************/
InsertTree (loc_tree, loc_r, loc_g, loc_b, loc_depth)
struct NODE       **loc_tree;
unsigned char     loc_r, loc_g, loc_b;
unsigned char     loc_depth;

{
  int i;           /* used for loops */

  /* if loc_tree points to NULL produce and init new octree */
  if (*loc_tree == NULL) {

    /* allocate memory for new node */
    if ( (*loc_tree = (struct NODE *) malloc ( (unsigned) sizeof (struct NODE)))
        == NULL ) {
      fprintf (stderr,"Error in malloc (tree)\n");
      free ( (char *) *loc_tree);
      exit(1);
    }

    /* init values of new node */
    (*loc_tree)->level = loc_depth;
    (*loc_tree)->leaf  = (loc_depth == global_octree_depth);
    (*loc_tree)->color_count = 0;
    (*loc_tree)->color_index = -1;
    (*loc_tree)->r = (*loc_tree)->g = (*loc_tree)->b = 0;

    if ( (*loc_tree)->leaf) {
      /* if new node is leaf then set size of tree */
      global_size++;
    }

    else {
      /* if new node is not leaf then init next pointers and */
      /* set the reduce list */
      for (i = 0; i < 8; i++) (*loc_tree)->next[i] = NULL;

      (*loc_tree)->next_node = global_reduce_list[loc_depth];
      global_reduce_list[loc_depth] = *loc_tree;
   }
  }

  if ((*loc_tree)->leaf) {
    /* if node is leaf then insert color in node */
    (*loc_tree)->color_count++;
    (*loc_tree)->r += loc_r;
    (*loc_tree)->g += loc_g;
    (*loc_tree)->b += loc_b;
  }

  /* if node was no leaf try to insert color in next node */
  else InsertTree (&((*loc_tree)->next[Branch(loc_r, loc_g, loc_b, loc_depth)]),
                   loc_r, loc_g, loc_b, loc_depth + 1);

}


/*******************************************************************************
 *   GetReducible                                                              *
 *                                                                             *
 *   Input  : - loc_tree  no input is taken                                    *
 *                                                                             *
 *   Output :                                                                  *
 *            - loc_tree  points to a reducible node                           *
 *                                                                             *
 ******************************************************************************/
GetReducible (loc_tree)
struct NODE   **loc_tree;

{
  struct NODE        *scan_node;
  unsigned long int  color_count;
  int                i;


  /* get list and fix depht */
  while (global_reduce_list[global_octree_depth - 1] == NULL)
    global_octree_depth--;

  /* if speed then take first node in list*/
  if (flag_opt_speed) {
    *loc_tree = global_reduce_list[global_octree_depth - 1];
    global_reduce_list[global_octree_depth - 1] =
      global_reduce_list[global_octree_depth - 1]->next_node;
  }


  else {

    /* fix color_count for all nodes in list */
    *loc_tree = global_reduce_list[global_octree_depth - 1];
    while ( (*loc_tree) != NULL) {
      color_count = 0;

      for (i = 0; i < 8; i++) {
        if ( (*loc_tree)->next[i] != NULL) 
          color_count += (*loc_tree)->next[i]->color_count;
       }
      (*loc_tree)->color_count = color_count;
    *loc_tree = (*loc_tree)->next_node;
    }

    /* if error then take node from list with smallest color_count */
    if (flag_opt_error) {
      /* last element in list ? */
      *loc_tree = global_reduce_list[global_octree_depth - 1];
      if ((*loc_tree)->next_node == NULL)
        global_reduce_list[global_octree_depth - 1] = NULL;

      else {
        /* swap the two first element if necessary */
        if ( (*loc_tree)->color_count <
          ((*loc_tree)->next_node)->color_count ) {
          global_reduce_list[global_octree_depth - 1] =
            (*loc_tree)->next_node;
          (*loc_tree)->next_node = ((*loc_tree)->next_node)->next_node;
          global_reduce_list[global_octree_depth - 1]->next_node = *loc_tree;
          *loc_tree = global_reduce_list[global_octree_depth - 1];
        }
        /* scan over all nodes in list */
        scan_node = *loc_tree;
        while ( (*loc_tree)->next_node != NULL) {
          if ( (scan_node->next_node)->color_count >
               ((*loc_tree)->next_node)->color_count )
            scan_node = *loc_tree;
          *loc_tree = (*loc_tree)->next_node;
        }
        /* get the node */
        *loc_tree = scan_node->next_node;
        scan_node->next_node = (scan_node->next_node)->next_node;
      }
    }

    /* if detail then take node from list with greatest color_count */
    else if (flag_opt_detail) {
      /* last element in list ? */
      *loc_tree = global_reduce_list[global_octree_depth - 1];
      if ((*loc_tree)->next_node == NULL)
        global_reduce_list[global_octree_depth - 1] = NULL;

      else {
        /* swap the two first element if necessary */
        if ( (*loc_tree)->color_count >
          ((*loc_tree)->next_node)->color_count ) {
          global_reduce_list[global_octree_depth - 1] =
            (*loc_tree)->next_node;
          (*loc_tree)->next_node = ((*loc_tree)->next_node)->next_node;
          global_reduce_list[global_octree_depth - 1]->next_node = *loc_tree;
          *loc_tree = global_reduce_list[global_octree_depth - 1];
        }

        /* scan over all nodes in list */
        scan_node = *loc_tree;
        while ( (*loc_tree)->next_node != NULL) {
          if ( (scan_node->next_node)->color_count <
               ((*loc_tree)->next_node)->color_count )
            scan_node = *loc_tree;
          *loc_tree = (*loc_tree)->next_node;
        }

        /* get the node */
        *loc_tree = scan_node->next_node;
        scan_node->next_node = (scan_node->next_node)->next_node;
      }

    }

    else {
      fprintf (stderr,"Illegal defaults set!\n");
      exit (1);
    }
  }
}


/*******************************************************************************
 *   ReduceTree                                                                *
 *                                                                             *
 *   Input  : - no input taken                                                 *
 *                                                                             *
 *   Output : - reduces with GetReducible the tree in one color                *
 *                                                                             *
 ******************************************************************************/
ReduceTree()
{

  struct NODE        *loc_tree;            /* pointer to node             */
  unsigned long int  sum_r, sum_g, sum_b;  /* sum of color values         */
  unsigned long int  color_count;          /* number of colors in node    */
  int                children;             /* number of childrens of node */
  int                i;                    /* used for loops              */

    /* init vars */
    loc_tree = NULL;
    sum_r = sum_g = sum_b = 0;
    color_count = 0;
    children = 0;

    /* find a reducible node */
    GetReducible (&loc_tree);

    /* set data of reducible node  */
    for (i = 0; i < 8; i++)

      if (loc_tree->next[i] != NULL) {
        /* add the colors of the children of node to the node himself */
        children++;
        color_count += loc_tree->next[i]->color_count;

        sum_r += loc_tree->next[i]->r;
        sum_g += loc_tree->next[i]->g;
        sum_b += loc_tree->next[i]->b;

        free ( (char *) loc_tree->next[i]);
      }


    /* the reduced node is now a leaf */
    loc_tree->leaf = TRUE;

    /* set color of node */
    loc_tree->color_count = color_count;
    loc_tree->r = sum_r;
    loc_tree->g = sum_g;
    loc_tree->b = sum_b;

    global_size = global_size - children + 1;
}


/*******************************************************************************
 *   InitColorTable                                                            *
 *                                                                             *
 *   Input  :                                                                  *
 *            - loc_tree  points to node                                       *
 *                                                                             *
 *   Output : - global colormap is set                                         *
 *                                                                             *
 ******************************************************************************/
InitColorTable (loc_tree)
struct NODE *loc_tree;

{
  int  i;    /* used for loops */

  if (loc_tree != NULL)
   if (loc_tree->leaf) {

      /* set colormap with the means of the leaf */
      global_red[global_map_index]   = (unsigned char) (loc_tree->r /
                                                     loc_tree->color_count);
      global_green[global_map_index] = (unsigned char) (loc_tree->g /
                                                     loc_tree->color_count);
      global_blue[global_map_index]  = (unsigned char) (loc_tree->b /
                                                     loc_tree->color_count);

      loc_tree->color_index = global_map_index;
      global_map_index++;
    }

    else for (i = 0; i < 8; i++) InitColorTable (loc_tree->next[i]);
}


/*******************************************************************************
 *   Quant                                                                     *
 *                                                                             *
 *   Input  :                                                                  *
 *            - loc_tree  pointer to node                                      *
 *            - loc_r     red value                                            *
 *            - loc_g     green value                                          *
 *            - loc_b     blue value                                           *
 *                                                                             *
 *   Output :                                                                  *
 *                                                                             *
 *   Returns the colorindex in colormap for color loc_r, loc_g, loc_b          *
 *                                                                             *
 ******************************************************************************/
Quant (loc_tree,loc_r, loc_g, loc_b)
struct NODE  *loc_tree;
unsigned char loc_r, loc_g, loc_b;

{
  if (loc_tree->leaf)
    return ( (unsigned char) loc_tree->color_index);
  else
    return (Quant (loc_tree->next[Branch (loc_r,loc_g,loc_b,loc_tree->level)],
                   loc_r, loc_g, loc_b));
}


/*******************************************************************************
 *   usage_exit                                                                *
 *                                                                             *
 *   Input  :                                                                  *
 *            - name      name of program as string                            *
 *                                                                             *
 *   Output :                                                                  *
 *            - writes the usage and the default values to stderr              *
 *                                                                             *
 *   Return : - exits with 1                                                   *
 *                                                                             *
 ******************************************************************************/
usage_exit (name)
char *name;
{

  fprintf (stderr,"\n%s V%s\n",name,VERSION);
  fprintf (stderr,"Usage: %s {options} <input filename> \n",name);
  fprintf (stderr,"               -o <output filename|%s> \n",
                                                              DEFAULT_STDOUT);
  fprintf (stderr,"               -en[+|-]       expand name with mode \n");
  fprintf (stderr,"               -ws[+|-]       write standard format \n");
  fprintf (stderr,"               -sv[+|-]       colors for sunview \n");
  fprintf (stderr,"               -opt speed     optimize speed \n");
  fprintf (stderr,"                |opt error    optimize error sum \n");
  fprintf (stderr,"                |opt detail   optimize detailed shadings\n");
  fprintf (stderr,"               -fs[+|-] [h d] use Floyd Steinberg ");
  fprintf (stderr,"dither\n");
  fprintf (stderr,"                                 following parameters ");
  fprintf (stderr,"in %% :\n");
  fprintf (stderr,"                                 h = horizontal error ");
  fprintf (stderr,"spreading\n");
  fprintf (stderr,"                                 d = diagonal error ");
  fprintf (stderr,"spreading\n");
  fprintf (stderr,"                                 h, d are integers ");
  fprintf (stderr,"(h+d<=100) \n");
  fprintf (stderr,"               -fs%% <0..100>  percent for Floyd ");
  fprintf (stderr,"Steinberg\n");
  fprintf (stderr,"                              dithering \n");
  fprintf (stderr,"               -ce[+|-]       calculate euclidean ");
  fprintf (stderr,"distance\n");
  fprintf (stderr,"                              and write to stdout \n");
  fprintf (stderr,"               -m[+|-]        write colormap to stdout \n");
  fprintf (stderr,"               -p[+|-]        prompt info to stderr \n");
  fprintf (stderr,"               -h             help (to stderr) \n");

  /* print the defaults */
  fprintf (stderr,"\nDefaults:                                       \n");
/* fprintf (stderr,"  - Extension of input file   : %s \n",DEFAULT_IN_EXT); */
  fprintf (stderr,"  - Extension of output file  : %s \n",DEFAULT_OUT_EXT);

  if (DEFAULT_EXP_NAM)
    fprintf (stderr,"  - Expand name               : + \n");
  else
    fprintf (stderr,"  - Expand name               : - \n");

  if (DEFAULT_OUT_FORM == RT_STANDARD)
    fprintf (stderr,"  - Write format              : standard \n");
  else
    fprintf (stderr,"  - Write format              : run length encoded \n");

  if (DEFAULT_SUNVIEW)
    fprintf (stderr,"  - Colors for sunview        : + \n");
  else
    fprintf (stderr,"  - Colors for sunview        : - \n");

  if (DEFAULT_FLOYD)
    fprintf (stderr,"  - Floyd Steinberg dither    : +  %d %d\n",
                        DEFAULT_SPREAD_HORI,DEFAULT_SPREAD_DIAG);
  else
    fprintf (stderr,"  - Floyd Steinberg dither    : - %d %d\n",
                       DEFAULT_SPREAD_HORI,DEFAULT_SPREAD_DIAG);

  fprintf (stderr,"  - Percent for Floyd Steinb. : %d \n",
           DEFAULT_SPREAD_PERCENT);

  if (DEFAULT_EUCLID)
    fprintf (stderr,"  - Calculate euclidean dist. : + \n");
  else
    fprintf (stderr,"  - Calculate euclidean dist. : - \n");

  if (DEFAULT_WRITE_MAP)
    fprintf (stderr,"  - Write colormap to stdout  : + \n");
  else
    fprintf (stderr,"  - Write colormap to stdout  : - \n");

  if (DEFAULT_PROMPT)
    fprintf (stderr,"  - Prompt info to stderr     : + \n");
  else
    fprintf (stderr,"  - Prompt info to stderr     : - \n");

  fprintf (stderr,"\n->try %s |& more \n",name);

  exit(1);
}


/*******************************************************************************
 *   help_exit                                                                 *
 *                                                                             *
 *   Input  :                                                                  *
 *            - name      name of program as string                            *
 *                                                                             *
 *   Output : - writes the help to stderr                                      *
 *                                                                             *
 *   Return : - exits with 1                                                   *
 *                                                                             *
 ******************************************************************************/
help_exit (name)
char *name;
{
  fprintf (stderr,"Help for %s V%s\n",name,VERSION);
  fprintf (stderr," I convert 24 or 32 Bit pixrect or rayshade images\n");
  fprintf (stderr," to 8 Bit colormapped images for suns using the octree\n");
  fprintf (stderr," algorithm. Here are some valid examples for calling:\n");
  fprintf (stderr,"   %s hsv.pixrect\n",name);
  fprintf (stderr,"   %s -sv -o %s tree.pic | rastool \n",name,DEFAULT_STDOUT);
  fprintf (stderr,"   %s -fs -o test.pic_8 test.pic_32 \n",name);
  fprintf (stderr,"   %s -fs -fs%% 75 -sv hsv\n",name);
  fprintf (stderr,"   %s -opt detail plant.pic\n",name);
  fprintf (stderr,"   %s -ce -opt error plant.pic\n,name");
  exit (1);
}





















